/*
 * Copyright 2024 Julien LE CLEACH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Supvisors common style for a card box
 */

div.card {
    padding: 0 15px;
    align-items: center;
    border: 1px solid black outset;
    border-radius: 6px;
    box-shadow: 0px 2px 10px black;
    background-image: linear-gradient(180deg, var(--dark1-color), var(--light2-color), var(--dark1-color));
}

div.card table.status {
    border: none;
    background: transparent;
}

div.card table.status th, div.card table.status td {
    border: none;
    background: transparent;
}

div.card table.status td {
    text-align: center;
}

div.card table.status .table_main {
    font-size: 1em;
    text-align: center;
}

div.card table.status .table_title {
    text-align: right;
}
