/*
 * Copyright 2024 Julien LE CLEACH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Supvisors Conciliation page
 */

/* Contents part */
#index_contents {
    display: flex;
    flex: 1;
    overflow: auto;
    padding: 10px;
}

/* update the background of the global strategies box */
div.strategies {
    background-image: linear-gradient(180deg, var(--light2-color), var(--light1-color));
}

div.strategies div.button_group {
    background-image: none;
}

div.strategies div.button_group:not(:first-child) {
    padding-top: 0;
}

/* update the area of the local strategies box */
div.local_strategies div {
    padding: 1px 3px;
}
