/*
 * Copyright 2024 Julien LE CLEACH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Supvisors Application page style
 */
 
/* Style for application operational status */
.status_empty, .status_green, .status_yellow, .status_red {
    padding: 15px 50px;
}

.status_empty {
    background: url(../img/leds_empty_30.png) no-repeat center;
}

.status_green {
    background: url(../img/leds_green_30.png) no-repeat center;
}

.status_yellow {
    background: url(../img/leds_yellow_30.png) no-repeat center;
}

.status_red {
    background: url(../img/leds_red_30.png) no-repeat center;
}

/* update the background of the strategies box */
#strategies {
    background-image: linear-gradient(180deg, var(--light2-color), var(--light1-color));
}

#strategies div.button_group {
    background-image: none;
}

#strategies div.button_group:nth-child(2) {
    padding-top: 0;
}

